IMPLEMENTATION MODULE Errors;

  (*========================================================================*)
  (* WICHTIG: BITTE NUR DIE DATEI FileSystem.def.m2pp EDITIEREN !!!         *)
  (*========================================================================*)
  (* Hint:                                                                  *)
  (* =====                                                                  *)
  (*                                                                        *)
  (* You  can create the definition file for the ISO M2 IO-channnel         *)
  (* module by "m2pp -D __ISO__ < Errors.mod.m2pp > Errors.mod              *)
  (*                                                                        *)
  (* For the original interface to M.Riedl IO-modules please use            *)
  (*           "m2pp -D __MRI__ < Errors.mod.m2pp > Errors.mod              *)
  (*                                                                        *)
  (*------------------------------------------------------------------------*)
  (* Fehlerbehandlungsmodul, M.Riedl, Nov. 1989                             *)
  (*                                                                        *)
  (* Module for error messages, M.Riedl, Nov. 1989                          *)
  (*                                                                        *)
  (* Last Change:                                                           *)
  (*                                                                        *)
  (*  Nov. 89, MRi: Erstellen der ersten Verions für JPI M2                 *)
  (* 12.06.95, MRi: Durchsicht                                              *)
  (* 09.11.14, MRi: Changed type conversion in WriteInt and WriteCard       *)
  (*                from LONGCARD(i) to VAL(LONGCARD,i)                     *)
  (* 05.02.16, MRi: WriteReal korrigiert.                                   *)
  (* 05.02.16, MRi: First ISO Modula-2 version                              *)
  (* 26.02.16, MRi: Locale now imported from Deklera                        *)
  (* 25.09.16, MRi: WriteBoolean eingefuegt                                 *)
  (*------------------------------------------------------------------------*)
  (* Licence : GNU Lesser General Public License (LGPL)                     *)
  (*------------------------------------------------------------------------*)

  (* $Id: Errors.mod.m2pp,v 1.4 2017/11/04 14:57:47 mriedl Exp mriedl $ *)

FROM Deklera  IMPORT Locale;

VAR    WarnStr : ARRAY [0..31] OF CHAR;
       ErrStr  : ARRAY [0..31] OF CHAR;


PROCEDURE SetLang(locale : ARRAY OF CHAR);

BEGIN
      IF    (locale[0] = "D") AND (locale[1] = "E") THEN
        Lang:=DE;
        WarnStr:=" *** WARNUNG : ";
        ErrStr :=" *** FEHLER : ";
      ELSIF (locale[0] = "E") AND (locale[1] = "N") THEN
        Lang:=EN;
        WarnStr:=" *** WARNING : ";
        ErrStr :=" *** ERROR : ";
      ELSIF (locale[0] = "F") AND (locale[1] = "R") THEN
        Lang:=FR;
        WarnStr:=" *** MISE EN GARDE: ";
        WarnStr:=" *** ERREUR : ";
      END;
END SetLang;

PROCEDURE Warnung(Fehlerflag : ARRAY OF CHAR);

          (*-----------------------------------------------------*)
          (* Schreibt Fehlerflag als Warnung auf den Bildschirm. *)
          (*-----------------------------------------------------*)
BEGIN
      WriteLn; WriteLn;
      WriteString(WarnStr); WriteString(Fehlerflag); WriteString(" ***");
      WriteLn; WriteLn;
END Warnung;

PROCEDURE FehlMeld(Fehlerflag : ARRAY OF CHAR);

          (*-----------------------------------------*)
          (* Schreibt Fehlerflag auf den Bildschirm. *)
          (*-----------------------------------------*)
BEGIN
      WriteLn; WriteLn;
      WriteString(ErrStr); WriteString(Fehlerflag); WriteString(" ***");
      WriteLn; WriteLn;
END FehlMeld;

PROCEDURE Exit(FFlag : ARRAY OF CHAR);

          (*-----------------------------------------------------------*)
          (* Schreibt Fehlerflag  auf den Bildschirm und schlie3t alle *)
          (* noch ge"offneten IO-Kan"ale (KEIN RETURN).                *)
          (*-----------------------------------------------------------*)

BEGIN
      FehlMeld(FFlag);
(*
      Sollte jetzt durch "FINALLY" in den Modulen abgefangen werden 
      Streams.CloseAll;      (* Verhindere Datenverluste ! *)
      FileSystem.CloseAll;
 *)
      HALT;
END Exit;

PROCEDURE ReadChar(VAR chr : CHAR); (* Lokal *)

BEGIN
END ReadChar;

PROCEDURE WriteChar(chr : CHAR);

BEGIN
END WriteChar;

PROCEDURE WriteString(Str : ARRAY OF CHAR);

          VAR i : CARDINAL;
BEGIN
      i:=0;
      LOOP
        IF (Str[i] = 0C) THEN EXIT END;
        WriteChar(Str[i]);
        INC(i);
        IF (i > HIGH(Str)) THEN EXIT END;
      END;
END WriteString;

PROCEDURE WriteLn;

BEGIN 
END WriteLn;

PROCEDURE WriteLongCard(c : LONGCARD);

          VAR i   : CARDINAL;
              s   : LONGCARD;
              Str : ARRAY [0..127] OF CHAR;
BEGIN
      i:=0;
      REPEAT
        s := c MOD 10; c := c DIV 10;
        Str[i] := CHR(s + VAL(LONGCARD,ORD("0"))); INC(i);
      UNTIL (c = 0);
(*
      UNTIL (c = 0) (* OR (i > HIGH(Str) *) ;
      IF (i > HIGH(Str) THEN
        RaiseException(...)
      END;
 *)        
      FOR s:=i-1 TO 0 BY -1 DO WriteChar(Str[s]); END;
END WriteLongCard;

PROCEDURE WriteCard(c : CARDINAL);
          
BEGIN
      WriteLongCard(VAL(LONGCARD,c));
END WriteCard;

PROCEDURE WriteInt(i : INTEGER);

BEGIN
       IF (i < 0) THEN WriteChar("-"); END;
       i := ABS(i);
       WriteLongCard(VAL(LONGCARD,i));
END WriteInt;

PROCEDURE WriteReal(x : LONGREAL);

          CONST Faktor = 1.0E+07;

          VAR   vor,nach : LONGCARD;
BEGIN
      IF (ABS(x) > LFLOAT(MAX(LONGCARD))) THEN
        WriteString("****.****");
        RETURN;
      ELSIF (x = 0.0) THEN
        vor:=0;
        nach:=0;
      ELSE
        IF (x < 0.0) THEN WriteChar("-"); x:=ABS(x); END;
        vor  := TRUNC(x);
        nach := TRUNC(Faktor*(x - LFLOAT(vor)));
      END;
      WriteLongCard(vor); WriteChar(".");
      WriteLongCard(nach);
END WriteReal;

PROCEDURE WriteBoolean(b : BOOLEAN);
BEGIN
       IF b THEN WriteString("TRUE"); ELSE WriteString("FALSE"); END;
END WriteBoolean;

PROCEDURE Pause(i : CARDINAL);

          VAR chr : CHAR;
BEGIN
      WriteLn; WriteString('Pause - '); WriteCard(i);
      ReadChar(chr); (* Warte bis eine Taste gedr"uckt wird. *)
      WriteLn; WriteLn;
END Pause;

BEGIN
      SetLang("DE");
      Fehler:=FALSE;
      Fehlerflag:=' ??? ';
      ErrOut:=Warnung;
      FatalError:=Exit;
END Errors.
