IMPLEMENTATION MODULE FileSystem;

  (*------------------------------------------------------------------------*)
  (* Implementation einer Dateisystemschittstelle f"ur Modula-2             *)
  (* (MtC, MOCKA, XDS M2, GM2).                                             *)
  (*                                                                        *)
  (* Letzte Aenderung:                                                      *)
  (*                                                                        *)
  (*       91, MRi: Erstellen der ersten Version fuer JPI M2                *)
  (* 13.05.93, MRi: Erstellen der ersten Version (MtC)                      *)
  (* 06.01.98, MRi: Anpassungen an Mocka.                                   *)
  (* 28.11.14, MRi: Anpassungen an XDS / GM2                                *)
  (* 08.01.15, MRi: Ersetzen des Aufrufs von stat durch fstat               *)
  (* 25.02.16, MRi: Einfuehrung der Prozedur IsOpen                         *)
  (* 31.10.16, MRi: RdBinObj & WrBinOnj eingefuehrt                         *)
  (*------------------------------------------------------------------------*)
  (* Licence : GNU Lesser General Public License (LGPL)                     *)
  (*------------------------------------------------------------------------*)

  (*
   * $Id: FileSystem.mod,v 1.2 2016/11/22 22:21:35 mriedl Exp mriedl $
   *)

FROM SYSTEM   IMPORT TSIZE,BYTE,ADR,CAST;
FROM UnixLib  IMPORT open,close,read,write,unlink,Stat,fstat,
                     chmod,sync,lseek,
                     LockExclusive,LockUnlock,LockNoblock,flock;
              IMPORT fslength;
FROM Storage  IMPORT ALLOCATE,DEALLOCATE;
              IMPORT StringsLib;
FROM Errors   IMPORT Fehler,Fehlerflag,ErrOut,FatalError;
              IMPORT Errors;

(*--------------------------- Lokale Objekte --------------------------------*)

TYPE  OpenModeRec         = RECORD (* Eingef"ugt 28.11.14 *)
                              CASE : INTEGER OF
                                | 1: openmode : OpenMode;
                                | 2: int      : INTEGER; (* avoid type cast *)
                              ELSE
                              END;
                            END;

      AccessModeRec       = RECORD 
                              CASE : INTEGER OF
                                | 1: accessmode : AccessMode;
                                | 2: int        : INTEGER; (* s.o. *)
                              ELSE
                              END;
                            END;

CONST MaxOpenFiles        =    36; (* _POSIX_OPEN_MAX *)
      MaxDateiNamenLaenge =   255; (* _POSIX_PATH_MAX *)
      SeekCur             =     2;
      Debug               = FALSE;
      UseStat             = TRUE;

VAR   DateiNamen : ARRAY [0..MaxOpenFiles-1] OF POINTER TO
                   ARRAY [0..MaxDateiNamenLaenge] OF CHAR;
      Geoeffnet  : ARRAY [0..MaxOpenFiles-1] OF BOOLEAN;

PROCEDURE StoreFileName(f      : File;           (* Kanalnummer *)
                        Name   : ARRAY OF CHAR); (* Dateiname   *)

                   (*------------------------------------------------*)
                   (* Speichert die Dateinamen aller ge\"offneten    *)
                   (* Dateien in dem globalen Zeigerfeld Dateinamen. *)
                   (*------------------------------------------------*)   

BEGIN
      IF (DateiNamen[f] # NIL) OR Geoeffnet[f] THEN
        IF Debug THEN
          Errors.WriteString("Kanalnummer (StoreFileName) : ");
          Errors.WriteCard(VAL(CARDINAL,f));
          IF Geoeffnet[f] THEN Errors.WriteString(" Kanal geoeffnet !"); END;
          IF DateiNamen[f] # NIL THEN 
            Errors.WriteString(" Dateiname nicht NIL !"); 
            Errors.WriteLn;
            Errors.WriteString(" Dateiname = "); 
            Errors.WriteString(DateiNamen[f]^);
            Errors.WriteLn;
          END;
          Errors.WriteLn;
        END;
        ErrOut("in FileSystem.StoreFileName !");
      END;
      NEW(DateiNamen[f]);
      IF (DateiNamen[f] = NIL) THEN
        FatalError("Kein Freispeicher vorhanden (FileSystem.StoreFileName) !");
      END;
      IF Debug THEN
        Errors.WriteString("StoreFileName : Datei (1) `"); Errors.WriteString(Name);
        Errors.WriteChar("'"); Errors.WriteLn;
      END;
      StringsLib.Copy(DateiNamen[f]^,Name);
      IF Debug THEN
        Errors.WriteString("StoreFileName : Datei (2) `"); 
        Errors.WriteString(DateiNamen[f]^);
        Errors.WriteChar("'"); Errors.WriteLn;
      END;
END StoreFileName;

PROCEDURE RemoveFileName(f : File); (* Kanalnummer *)

BEGIN
      IF (DateiNamen[f] = NIL) OR (DateiNamen[f]^[0] = 0C) THEN
        Fehlerflag:=
        "Fehler in FileSystem.RemoveFileName !";
        IF CheckIO THEN ErrOut(Fehlerflag); END;
        Fehler:=TRUE; RETURN;
      ELSE
        IF Debug THEN
          Errors.WriteString("Dateiname #"); Errors.WriteString(DateiNamen[f]^);
          Errors.WriteChar("#"); Errors.WriteLn;
        END;
        DISPOSE(DateiNamen[f]);
        DateiNamen[f]:=NIL; (* MR, 15.02.2015 *)
      END;
END RemoveFileName;

PROCEDURE IstZulaessigeKanalnummer(f : File) : BOOLEAN;

BEGIN
      RETURN NOT ((f < 0) OR (f >= MaxOpenFiles));
END IstZulaessigeKanalnummer;

(*--------------------------- Globale Objekte --------------------------------*)

PROCEDURE GetFileName(    f    : File;           (* Kanalnummer *)
                      VAR Name : ARRAY OF CHAR); (* Dateiname   *)

BEGIN
      IF NOT Geoeffnet[f] THEN
        Name[0] := 0C;
      ELSE
        StringsLib.Copy(Name,DateiNamen[f]^);
      END;
END GetFileName;

PROCEDURE IsOpen(f : File) : BOOLEAN;

BEGIN 
      IF (f < MaxOpenFiles) THEN
        RETURN Geoeffnet[f];
      ELSE
        RETURN FALSE;
      END;
END IsOpen;

PROCEDURE Lookup(VAR f        : File;
                     Name     : ARRAY OF CHAR;
                     creat    : BOOLEAN;
                     OpenMask : OpenMode);

          VAR DName : ARRAY [0..MaxDateiNamenLaenge] OF CHAR;
              err   : INTEGER;
              OM    : OpenModeRec;
              OMloc : OpenModeRec;

BEGIN
      Fehler := FALSE;
      IF (StringsLib.Length(Name) = 0) THEN
        Fehlerflag:="Dateiname leer (FileSystem.Lookup) !";
        IF CheckIO THEN ErrOut(Fehlerflag); END;
        Fehler:=TRUE; f:=-1;
        RETURN;
      END;
      StringsLib.Copy(DName,Name); (* Wegen m"oglicherweise fehlendem 0C *)

      (* "Offnen oder Erzeugen der Datei. *)

      OM.openmode:=OpenMask;
      IF creat THEN
     
(*?     f := open(ADR(DName),CAST(INTEGER,OpenMode{(*openRDONLY*)}));
 *)     OMloc.int:=0;
        INCL(OMloc.openmode,openRDONLY);
        f := open(ADR(DName),OMloc.int);
        IF (f # -1) THEN (* Datei vorhanden, also L"oschen. *)
          f := close(f); 
          f := unlink(ADR(DName));
          IF (f = -1) THEN (* Datei nicht zu l"oschen ! *)
            Fehler:=TRUE;
            Fehlerflag:="Zugriff verweigert (FileSystem.Lookup) !";
            fslength.SetErrFlag(Fehlerflag);
            IF CheckIO THEN ErrOut(Fehlerflag); END;
            RETURN;
          END;
        END; (* Datei (neu) anlegen. *)
        INCL(OM.openmode,openCREAT);
        (* Shift von OpenMask wegen "uberz"ahligem Bit *)
(* 28.11.2014
        OpenMask := OpenMode((CARDINAL(OpenMask) + 0) DIV 2);
 *)
        OMloc.int := ((OM.int + 0) DIV 2);
(*?*)   f   := open(ADR(DName),OMloc.int);
        err := chmod(ADR(DName),CAST(INTEGER,defaultAccess));
      ELSE (* "Offnen zum Lesen, Schreiben etc. *)
        OM.int := ((OM.int + 0) DIV 2);
(*?*)   f := open(ADR(DName),OM.int);
      END;

      IF (f = -1) THEN
        IF creat THEN
          Fehlerflag:="Fehler bei Erstellen der Datei '";
        ELSE
          Fehlerflag:="Fehler bei Oeffnen der Datei '";
        END;
        StringsLib.Append(Fehlerflag,DName);
        StringsLib.Append(Fehlerflag,"' (FileSystem.Lookup) !");
        IF CheckIO THEN ErrOut(Fehlerflag); END;
        Fehler:=TRUE; RETURN;
      ELSIF (f >= MaxOpenFiles) THEN
        Fehlerflag:="Zuviele geoeffnete Dateien (FileSystem.Lookup) !";
        IF CheckIO THEN ErrOut(Fehlerflag); END;
        Fehler:=TRUE; f:=-1; RETURN;
      ELSE (* Eintagen in die Liste der ge"offneten Dateien. *)
        StoreFileName(f,DName);
        Geoeffnet[f] := TRUE;
        (* Sperren gegen Beschreiben durch andere Prozesse. *)
        (* lockf(f,1,MAX(LONGINT)); *)
      END;
END Lookup;

PROCEDURE ReWrite(VAR f : File);

          VAR DName : ARRAY [0..MaxDateiNamenLaenge] OF CHAR;

BEGIN
      Fehler:=FALSE;
      IF NOT Geoeffnet[f] THEN
        Fehlerflag:=" Datei nicht geoeffnet (FileSystem.ReWrite) ";
        IF CheckIO THEN ErrOut(Fehlerflag); END;
        Fehler := TRUE; f:=-1; RETURN;
      END;

      GetFileName(f,DName); (* Dateiname von f ermitteln, dann *)
      IF Debug THEN
        Errors.WriteString("ReWrite, Dateiname : "); Errors.WriteString(DName);
        Errors.WriteChar("#"); Errors.WriteLn;
        Errors.WriteString("ReWrite, Kanalnummer : "); Errors.WriteInt(f);
        Errors.WriteLn;
      END;

      Close(f);                 (* Schlie3en und neu"offnen mit Abschneiden. *)

      (* Hier fehlt die ermittlung, in welchem Modus die Datei urspr"unglich *)
      (* ge"offnet wurde (z.B. openRDONLY).                                  *)

(*?*) f := open(ADR(DName),(CAST(INTEGER,OpenMode{openRDWR,openTRUNC}) + 0) DIV 2);

      IF Debug THEN
        Errors.WriteString("ReWrite, Kanalnummer (neu) : ");
        Errors.WriteInt(f); Errors.WriteLn;
      END;
      IF (f = -1) THEN
        Fehlerflag:=" Dateioeffnungsfehler (FileSystem.ReWrite) !";
        IF CheckIO THEN ErrOut(Fehlerflag); END;
        Fehler:=TRUE; RETURN;
      ELSIF (f >= MaxOpenFiles) THEN
        Fehlerflag:="Zuviele geoeffnete Dateien (FileSystem.ReWrite) !";
        IF CheckIO THEN ErrOut(Fehlerflag); END;
        Fehler:=TRUE; f:=-1; RETURN;
      ELSE (* Eintagen in die Liste der ge"offneten Dateien. *)
        StoreFileName(f,DName);
        Geoeffnet[f] := TRUE;
        (* Sperren gegen Beschreiben durch andere Prozesse. *)
        (* lockf(f,1,MAX(LONGINT)); *)
      END;
      IF Debug THEN
        Errors.WriteString("ReWrite, Dateiname : #"); Errors.WriteString(DName); 
        Errors.WriteChar("#"); Errors.WriteLn; 
        Errors.WriteString("Dateiname (eingetragener) : `");
        Errors.WriteString(DateiNamen[f]^); Errors.WriteChar("'"); Errors.WriteLn;
      END;
END ReWrite;

PROCEDURE Exists(DName : ARRAY OF CHAR) : BOOLEAN;

          VAR datei   : File;
              interim : BOOLEAN;
BEGIN
      interim:=CheckIO;
      CheckIO:=FALSE;
      Lookup(datei,DName,FALSE,OpenReadOnly);
      CheckIO:=interim; 
      IF (datei = -1) THEN
        RETURN FALSE;
      ELSE
        Close(datei);
        RETURN TRUE;
      END;
END Exists;

PROCEDURE Close(VAR f : File);

BEGIN
      IF NOT Geoeffnet[f] THEN
        Fehlerflag:=" Datei nicht geoeffnet (FileSystem.Close) !";
        IF CheckIO THEN ErrOut(Fehlerflag); END;
        Fehler:=TRUE; f:=-1;
        RETURN;
      END;
      RemoveFileName(f);     (* Reihenfolge beachten ! *)
      Geoeffnet[f] := FALSE;
      (* unlockf(f); (* Sperre freigeben *) *)
      f := close(f);
      IF (f < 0) THEN
        ErrOut("Fehler in FileSystem.Close !");
      END;
END Close;

PROCEDURE CloseAll();

          VAR f : File;
              i : INTEGER;
BEGIN
      FOR i:=3 TO MaxOpenFiles-1 DO
        f:=VAL(File,i);
        IF Geoeffnet[f] THEN Close(f); END; 
      END;
END CloseAll;

PROCEDURE Delete(    Name : ARRAY OF CHAR;
                 VAR f    : File);

          VAR DName : ARRAY [0..MaxDateiNamenLaenge] OF CHAR;

BEGIN
      Fehler:=FALSE;

      GetFileName(f,DName);
      IF (StringsLib.Compare(DName,Name) # StringsLib.equal) OR NOT Geoeffnet[f] THEN
        Fehlerflag:=" Datei nicht geoeffnet (FileSystem.Delete) !";
        IF CheckIO THEN ErrOut(Fehlerflag); END;
        Fehler:=TRUE; RETURN;
      ELSE
        Close(f);
        f := unlink(ADR(Name));
        IF (f = -1) THEN
          Fehlerflag:=" Zugriff verweiget (FileSystem.Delete) !";
          IF CheckIO THEN ErrOut(Fehlerflag); END;
          Fehler:=TRUE; RETURN;
        END;
      END;
END Delete;

PROCEDURE Length(    f   : File;
                 VAR Len : LONGCARD);

          VAR StatPuf : Stat;
              DName   : ARRAY [0..MaxDateiNamenLaenge] OF CHAR;
              error   : INTEGER;

BEGIN
      Fehler:=FALSE;
      IF NOT IstZulaessigeKanalnummer(f) THEN
        FatalError(" Unzulassige Kanalnummer (FileSystem.Length) !");
      END;
      IF NOT Geoeffnet[f] THEN
        Fehlerflag:=" Datei nicht geoeffnet (FileSystem.Length) !";
        IF CheckIO THEN ErrOut(Fehlerflag); END;
        Fehler:=TRUE; RETURN;
      END;
      IF NOT UseStat THEN (* Viel zu kompliziert, fstat ist einfacher *)
        GetFileName(f,DName); (* Dateiname von f ermitteln, dann *)
        fslength.FileLength(Len,error,DName);
      ELSE
        (* error := stat(ADR(DName),StatPuf); *)
        error := fstat(f,StatPuf);
Errors.Pause(777);
      END;
      IF (error = -1) THEN
        (* Fehlerflag := strerror(errno); *)
        Fehlerflag:="Dateilaenge nicht ermittelbar (FileSystem.Length) !"; 
        IF CheckIO THEN ErrOut(Fehlerflag); END;
        Fehler:=TRUE; Len:=MAX(LONGCARD);
      END;
      IF UseStat THEN
Errors.Pause(888);
        Len := StatPuf.stSize;
Errors.Pause(999);
      END;
END Length;

PROCEDURE GetPos(    f   : File;
                 VAR Pos : LONGCARD);

          VAR IPos : LONGINT;

BEGIN
      Fehler:=FALSE;
      IPos := VAL(LONGINT,lseek(f,0,SeekCur));
      IF (IPos = -1) THEN
        Fehlerflag:=" Fehler in FileSystem.GetPos";
        IF CheckIO THEN ErrOut(Fehlerflag); END;
        Fehler:=TRUE; Pos:=MAX(LONGCARD);
        RETURN;
      END;
      Pos := VAL(LONGCARD,IPos);
END GetPos;

PROCEDURE SetPos(VAR f   : File;
                     Pos : LONGCARD);

          VAR IPos : LONGINT;

BEGIN
      Fehler:=FALSE;
      IF NOT Geoeffnet[f] THEN
        Fehlerflag:=" Datei nicht geoeffnet (FileSystem.SetPos) !";
        IF CheckIO THEN ErrOut(Fehlerflag); END;
        Fehler:=TRUE; RETURN;
      END;
      IPos := VAL(LONGINT,lseek(f,Pos,0));
      IF (IPos = -1) THEN
        Fehlerflag:=" SetPos nicht moeglich (FileSystem.SetPos) !";
        IF CheckIO THEN ErrOut(Fehlerflag); END;
        Fehler:=TRUE;
      END;
END SetPos;

PROCEDURE ReadNBytes(    f    : File;
                     VAR Ziel : ARRAY OF BYTE;
                         n    : CARDINAL;
                     VAR m    : CARDINAL);

          VAR nbyte : INTEGER;
BEGIN
      nbyte := read(f,ADR(Ziel),n);
      IF (nbyte = -1) THEN
        fslength.SetErrFlag(Fehlerflag);
        IF CheckIO THEN ErrOut(Fehlerflag); END;
        Fehler:=TRUE; m:=0; 
        RETURN;
      END;
      m := VAL(CARDINAL,ABS(nbyte));
END ReadNBytes;

PROCEDURE WriteNBytes(    f      : File;
                      VAR Quelle : ARRAY OF BYTE;
                          n      : CARDINAL;
                      VAR m      : CARDINAL);

          VAR nbytes : INTEGER;
BEGIN
      Fehler:=FALSE;
      nbytes := write(f,ADR(Quelle),n);
      IF (nbytes = -1) THEN
        Fehlerflag:=" Schreiben nicht moeglich (FileSystem.WriteNBytes) !";
        IF CheckIO THEN ErrOut(Fehlerflag); END;
        Fehler:=TRUE; m := 0; RETURN;
      END;
      m := VAL(CARDINAL,nbytes);
END WriteNBytes;

PROCEDURE RdBinObj(    Ein   : File;
                   VAR A     : ARRAY OF BYTE;
                       N     : CARDINAL;
                   VAR ifehl : INTEGER);

       CONST BlockSize = 16*1024;
             BlockLen  = BlockSize DIV TSIZE(BYTE);

       VAR   NBytes,NBlocks,rest : CARDINAL;
             iblk,ii,nbytes      : CARDINAL;
BEGIN
       ifehl:=0;
       IF (N > HIGH(A)+1) THEN
         ifehl:=2;
         Fehler:= TRUE; Fehlerflag:="Feld A zu klein (RdBinObj)";
         ErrOut(Fehlerflag);
         RETURN;
       END;
       NBytes  := N*TSIZE(BYTE);
       NBlocks := NBytes DIV BlockSize;
       rest    := NBytes MOD BlockSize;
       ii:=0;
       IF (NBlocks > 0) THEN
         iblk:=1;
         REPEAT
           ReadNBytes(Ein,(*ADR*)A[ii],BlockSize,nbytes);
           Fehler:= (nbytes # BlockSize);
           INC(iblk); INC(ii,BlockLen); 
         UNTIL (iblk > NBlocks) OR Fehler;
       END;
       IF NOT Fehler AND (rest > 0) THEN
         ReadNBytes(Ein,(*ADR*)A[ii],rest,nbytes);
         Fehler:= (nbytes # rest);
       END;
       IF Fehler THEN
         ifehl:=1;
         Fehlerflag:="Keine ausreichende Zahl Bytes einlesbar (RdBinObj)";
         ErrOut(Fehlerflag);
       END;
END RdBinObj;

PROCEDURE WrBinObj(    Aus   : File;
                   VAR A     : ARRAY OF BYTE;
                       N     : CARDINAL;
                   VAR ifehl : INTEGER);

       CONST BlockSize = 16*1024;
             BlockLen  = BlockSize DIV TSIZE(BYTE);

       VAR   NBytes,NBlocks,rest : CARDINAL;
             iblk,ii,nbytes      : CARDINAL;
BEGIN
       ifehl:=0;
       IF (N > HIGH(A)+1) THEN
         ifehl:=2;
         Fehler:= TRUE; Fehlerflag:="Feld A zu klein (WrBinObj)";
         ErrOut(Fehlerflag);
         RETURN;
       END;
       NBytes  := N*TSIZE(BYTE);
       NBlocks := NBytes DIV BlockSize;
       rest    := NBytes MOD BlockSize;
       ii:=0;
       IF (NBlocks > 0) THEN
         iblk:=1;
         REPEAT
           WriteNBytes(Aus,(*ADR*)A[ii],BlockSize,nbytes);
           Fehler:= (nbytes # BlockSize);
           INC(iblk); INC(ii,BlockLen); 
         UNTIL (iblk > NBlocks) OR Fehler;
       END;
       IF NOT Fehler AND (rest # 0) THEN
         WriteNBytes(Aus,(*ADR*)A[ii],rest,nbytes);
         Fehler:= (nbytes # rest);
       END;
       IF Fehler THEN
         ifehl:=1;
         Fehlerflag:="Keine ausreichende Zahl Bytes einlesbar (WrBinObj)";
         ErrOut(Fehlerflag);
       END;
END WrBinObj;

          (*============== Unix - spezifische Routinen. ==============*)

PROCEDURE ChMod(VAR f     : File;
                    Modus : AccessMode);

          VAR err   : INTEGER;
              DName : ARRAY [0..MaxDateiNamenLaenge] OF CHAR;
              Mode  : AccessModeRec;

BEGIN
      Fehler:=FALSE;
      IF NOT Geoeffnet[f] THEN
        Fehlerflag:=" Datei nicht geoeffnet (FileSystem.ChMod) !";
        IF CheckIO THEN ErrOut(Fehlerflag); END;
        Fehler:=TRUE; f:=-1; RETURN;
      END;
      GetFileName(f,DName); (* Dateiname von f ermitteln, dann *)
      Mode.accessmode:=Modus;
      err := chmod(ADR(DName),Mode.int);
      IF (err = -1) THEN (* Zugriff verweigert ? *)
        Fehlerflag:=" Datei nicht gefunden (FileSystm.ChMod) !";
        IF CheckIO THEN ErrOut(Fehlerflag); END;
        Fehler:=TRUE; RETURN;
      END;
END ChMod;

PROCEDURE Lock(VAR f : File);

          VAR err,lex,lnb,lock : INTEGER;
              tmp              : BITSET;
BEGIN
      (* Folgendes Konstrukt ist n"otig, da CAST(TYPE,CONST) bei einigen *)
      (* "Ubersetzern nicht m"oglich.                                    *)
      lex := LockExclusive;
      lnb := LockNoblock;
      tmp := CAST(BITSET,lex) + CAST(BITSET,lnb);
      lock := CAST(INTEGER,tmp);
      err := flock(f,lock);
      IF (err # 0) THEN
        fslength.SetErrFlag(Fehlerflag);
        ErrOut(Fehlerflag);
      END; 
END Lock;

PROCEDURE UnLock(VAR f : File);

          VAR err : INTEGER;
BEGIN
      err := flock(f,LockUnlock);
      IF (err # 0) THEN
        fslength.SetErrFlag(Fehlerflag);
        ErrOut(Fehlerflag);
      END; 
END UnLock;

PROCEDURE Sync();

BEGIN
      sync();
END Sync;

      VAR i     : CARDINAL;
      (*  dummy : File; *)
BEGIN
      FOR i:=0 TO MaxOpenFiles-1 DO 
        Geoeffnet[i]:=FALSE;
        DateiNamen[i]:=NIL; 
      END;

(*    Lookup(dummy,"con"   ,FALSE);  ChMod(dummy,Permission(ReadWrite)); *)

      Input :=0;
      Output:=1;
      StdErr:=2;
      NEW(DateiNamen[Input ]);
      NEW(DateiNamen[Output]);
      NEW(DateiNamen[StdErr]);
      Geoeffnet[Input ] := TRUE; DateiNamen[Input ]^ := "Input";
      Geoeffnet[Output] := TRUE; DateiNamen[Output]^ := "Output";
      Geoeffnet[StdErr] := TRUE; DateiNamen[StdErr]^ := "StdErr";

      (* Wenn TRUE, kann es zu kaskadierenden Fehermeldungen im Modul *)
      (* Streams kommen.                                              *)

      CheckIO  := TRUE; 

FINALLY

      CheckIO  := FALSE;
      CloseAll();

END FileSystem.
