IMPLEMENTATION MODULE SFormEin;

  (*========================================================================*)
  (* WICHTIG: BITTE NUR DIE DATEI FileSystem.def.m2pp EDITIEREN !!!         *)
  (*========================================================================*)
  (* Hint:                                                                  *)
  (* =====                                                                  *)
  (*                                                                        *)
  (* You  can create the definition file for the ISO M2 IO-channnel         *)
  (* module by "m2pp -D __ISOIO__ < SFormEin.mod.m2pp > SFormEin.mod        *)
  (*                                                                        *)
  (* For the original interface to M.Riedl IO-modules please use            *)
  (*           "m2pp -D __MRI__ < SFormEin.mod.m2pp > SFormEin.mod          *)
  (*                                                                        *)
  (*------------------------------------------------------------------------*)
  (* Modul fuer das formatierte Einlesen von Daten - dient der Simulation   *)
  (* von FORTRAN "formatted input"                                          *)
  (*------------------------------------------------------------------------*)
  (* Letzte Bearbeitung                                                     *)
  (*                                                                        *)
  (* 27.11.93, MRi: Durchsicht                                              *)
  (* 04.11.17, MRi: Anpassung an ISO chanId                                 *)
  (*                                                                        *)
  (*------------------------------------------------------------------------*)
  (* Licence : GNU Lesser General Public License (LGPL)                     *)
  (*------------------------------------------------------------------------*)

  (* $Id: SFormEin.mod.m2pp,v 1.2 2017/11/04 15:54:27 mriedl Exp mriedl $ *)

                 IMPORT ASCII;
FROM Errors      IMPORT Fehlerflag,ErrOut,FatalError;
FROM ConvTypes   IMPORT ConvResults;
FROM CommonIO    IMPORT EoL,EoF;
FROM StringsLib  IMPORT Length,Concat;
FROM Streams     IMPORT Stream;
FROM NumConvert  IMPORT StringToCard;
FROM RealConvert IMPORT StringToReal;
                 IMPORT SIO;

VAR   Init   : BOOLEAN; (* EOL auf FALSE gesetzt ? *)
      result : ConvResults;
      FFlag  : ARRAY [strAllRight..strEmpty] OF ARRAY [0..31] OF CHAR;

PROCEDURE RdWort(VAR Ein    : Stream;
                 VAR Wort   : ARRAY OF CHAR;
                     Breite : CARDINAL;
                 VAR Null   : BOOLEAN);

          (*--------------------------------------------------------------*)
          (* Liest eine Zeichenkette der Breite Breite von der Datei Ein. *)
          (* Null = TRUE, wenn keine Zeichenkette gelesen wurde           *)
          (*--------------------------------------------------------------*)

          VAR   Chr      : CHAR;
                j,k      : CARDINAL;

BEGIN
      Null:=FALSE; k:=0;
      REPEAT  (* "Uberlese Leerzeichen *)
        INC(k); SIO.RdChar(Ein,Chr);
        IF (Chr = EoF) OR (Chr = EoL) OR ((k = Breite) AND (Chr = ' ')) THEN
          Null:=TRUE; Wort[0]:=0C; RETURN;
        END;
      UNTIL (Chr # ' ');
      Wort[0]:=Chr; (* Lese das Wort *)
      IF (k = Breite) THEN Wort[1]:=0C; RETURN; END;
      j:=0;
      LOOP  (* Lese bis EoL,EoF,Leerzeichen. *)
        INC(j); INC(k); SIO.RdChar(Ein,Chr);
        IF (Chr = ' ') OR (Chr = EoL) OR (Chr = EoF) THEN EXIT; END;
        Wort[j]:=Chr;
        IF (k = Breite) THEN INC(j); EXIT END; (* Richtiger Abschluss. *)
      END; (* LOOP *)
      Wort[j]:=0C;
      WHILE (k < Breite) AND NOT Ein.EOL DO (* Lese bis Ende des Felds *)
        INC(k); SIO.RdChar(Ein,Chr);
      END;
END RdWort;

PROCEDURE RdLn(VAR Ein : Stream);

          VAR   Chr : CHAR;
BEGIN
      IF NOT Init THEN Init:=TRUE; Ein.EOL:=FALSE; Ein.EOF:=FALSE; END;
      IF Ein.EOF THEN RETURN END;
      IF Ein.EOL AND NOT Ein.EOF THEN
        Ein.EOL:=FALSE;
      ELSE
        REPEAT SIO.RdChar(Ein,Chr); UNTIL Ein.EOL OR Ein.EOF;
        IF (NOT Ein.EOF) THEN Ein.EOL:=FALSE; END;
      END;
END RdLn;

PROCEDURE RdItemF(VAR Ein    : Stream;
                  VAR Str    : ARRAY OF CHAR;
                      Breite : CARDINAL);

           VAR i,j,Len : CARDINAL;
               Null    : BOOLEAN;
               String  : ARRAY [0..127] OF CHAR;

BEGIN
        IF (HIGH(Str) < Breite) THEN
          FatalError('HIGH(Str) < Breite (SFormEin.RdItermF).');
        END;
        RdWort(Ein,String,Breite,Null);
        Len:=Length(String);
        IF (Len < Breite) THEN
          IF (Len > 0) THEN FOR i:=0 TO Len-1 DO Str[i]:=String[i]; END; END;
          FOR i:=Len TO Breite - 1 DO Str[i]:=' '; END;
(*
          FOR i:=0 TO (Breite - Len - 1) DO Str[i]:=' '; END;
          j:=0;
          FOR i:=(Breite - Len) TO (Breite - 1) DO
            Str[i]:=String[j]; INC(j);
          END;
*)
        END;
        Str[Breite]:=0C;
END RdItemF;

PROCEDURE RdStrF(VAR Ein    : Stream;
                 VAR Str    : ARRAY OF CHAR;
                     Breite : CARDINAL);

          VAR Null : BOOLEAN;

BEGIN
      RdWort(Ein,Str,Breite,Null);
END RdStrF;

PROCEDURE RdCardF(VAR Ein    : Stream;
                  VAR c      : CARDINAL;
                      Breite : CARDINAL);

          CONST MaxCardStr = 9;

          VAR   CardStr    : ARRAY [0..MaxCardStr] OF CHAR;
                Null,OK    : BOOLEAN;

BEGIN
      IF NOT Init THEN Init:=TRUE; Ein.EOL:=FALSE; END;
      IF (Breite > MaxCardStr) THEN
        ErrOut('Breite zu gro\3 (SFormEin.RdCardF)');
        Breite:=MaxCardStr;
      END;
      IF Ein.EOL THEN
        c:=0; RETURN;
      ELSE
        RdWort(Ein,CardStr,Breite,Null);
      END;
      IF Null THEN
        c:=0;
      ELSE
        StringToCard(CardStr,c,result);
        IF (result # strAllRight) THEN
          Concat(Fehlerflag,FFlag[result]," (SFormEin.RdCardF.) !");
          ErrOut(Fehlerflag);
          c:=MAX(CARDINAL);
        END;
      END;
END RdCardF;

PROCEDURE RdLngRealF(VAR Ein    : Stream;
                     VAR x      : LONGREAL;
                         Breite : CARDINAL);

          CONST  MaxRealStr = 30;

          VAR    RealStr    : ARRAY [0..MaxRealStr] OF CHAR;
                 OK,Null    : BOOLEAN;

BEGIN
      IF NOT Init THEN Init:=TRUE; Ein.EOL:=FALSE; END;
      IF (Breite > MaxRealStr) THEN
        ErrOut('Breite zu gro\3 (SFormEin.RdLngRealF)');
        Breite:=MaxRealStr;
      END;
      IF Ein.EOL THEN
        x:=0.0; RETURN;
      ELSE
        RdWort(Ein,RealStr,Breite,Null);
      END;
      IF Null THEN
        x:=0.0;
      ELSE
        StringToReal(RealStr,x,result);
        IF (result # strAllRight) THEN
          Concat(Fehlerflag,FFlag[result]," (SFormEin.RdCardF.) !");
          ErrOut(Fehlerflag);
          x:=MAX(LONGREAL);
        END;
      END;
END RdLngRealF;

PROCEDURE WrStrF(VAR Aus    : Stream;
                     Str    : ARRAY OF CHAR;
                     Breite : CARDINAL;
                     Form   : CARDINAL);

          VAR Len,i : CARDINAL;

BEGIN
      Len:=Length(Str);
      IF (Len > Breite) THEN Str[Breite]:=0C; Len:=Breite; END;
      IF (Form = 0) THEN
        FOR i:=1 TO Breite - Len DO SIO.WrChar(Aus,' '); END;
        SIO.WrStr(Aus,Str);
      ELSE
        SIO.WrStr(Aus,Str);
        FOR i:=1 TO Breite - Len DO SIO.WrChar(Aus,' '); END;
      END;
END WrStrF;

BEGIN
      Init:=FALSE;
      FFlag[strAllRight]    := 'In Ordnung !';
      FFlag[strOutOfRange]  := 'Wertebereichs"uberschreitung !';
      FFlag[strWrongFormat] := 'Falsches Format !';
      FFlag[strEmpty]       := 'Zeichenkette leer !';
END SFormEin.
